import asyncio
from datetime import datetime
from pathlib import Path
from typing import Dict

import rerun as rr
from loguru import logger
from nova import MotionGroup
from nova.actions import Action, CombinedActions
from nova.api import models
from nova.core.nova import Nova
from wandelbots_api_client.models import (
    FeedbackOutOfWorkspace,
    PlanTrajectoryFailedResponseErrorFeedback,
)

from nova_rerun_bridge import colors
from nova_rerun_bridge.blueprint import send_blueprint
from nova_rerun_bridge.collision_scene import log_collision_scenes
from nova_rerun_bridge.consts import RECORDING_INTERVAL, TIME_INTERVAL_NAME
from nova_rerun_bridge.helper_scripts.download_models import get_project_root
from nova_rerun_bridge.stream_state import stream_motion_group
from nova_rerun_bridge.trajectory import TimingMode, log_motion


class NovaRerunBridge:
    """Bridge between Nova and Rerun for visualization.

    This class provides functionality to visualize Nova data in Rerun.
    It handles trajectoy, collision scenes, blueprints and proper cleanup of resources.

    Example:
        ```python
        from nova.core.nova import Nova
        from nova_rerun_bridge import NovaRerunBridge

        async def main():
            nova = Nova()
            async with NovaRerunBridge(nova) as bridge:
                await bridge.setup_blueprint()
                await bridge.log_collision_scenes()

        ```

    Args:
        nova (Nova): Instance of Nova client
        spawn (bool, optional): Whether to spawn Rerun viewer. Defaults to True.
    """

    def __init__(self, nova: Nova, spawn: bool = True, recording_id=None) -> None:
        self._ensure_models_exist()
        self.nova = nova
        self._streaming_tasks = {}
        if spawn:
            recording_id = recording_id or f"nova_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
            rr.init(application_id="nova", recording_id=recording_id, spawn=True)
        logger.add(sink=rr.LoggingHandler("logs/handler"))

    def _ensure_models_exist(self):
        """Ensure robot models are downloaded"""
        models_dir = Path(get_project_root()) / "models"
        if not models_dir.exists() or not list(models_dir.glob("*.glb")):
            print("Models not found, run update_robot_models() or poetry run download-models")

    async def setup_blueprint(self) -> None:
        """Configure and send blueprint configuration to Rerun.

        Fetches motion groups from Nova and configures visualization layout.
        """
        cell = self.nova.cell()

        controllers = await cell.controllers()
        motion_groups = []

        if not controllers:
            logger.warning("No controllers found")
            return

        for controller in controllers:
            for motion_group in await controller.activated_motion_groups():
                motion_groups.append(motion_group.motion_group_id)

        send_blueprint(motion_groups)

    async def log_collision_scenes(self) -> Dict[str, models.CollisionScene]:
        """Fetch and log all collision scenes from Nova to Rerun."""
        collision_scenes = (
            await self.nova._api_client.store_collision_scenes_api.list_stored_collision_scenes(
                cell=self.nova.cell()._cell_id
            )
        )
        log_collision_scenes(collision_scenes)
        return collision_scenes

    async def log_collision_scene(self, scene_id: str) -> Dict[str, models.CollisionScene]:
        """Log a specific collision scene by its ID.

        Args:
            scene_id (str): The ID of the collision scene to log

        Raises:
            ValueError: If scene_id is not found in stored collision scenes
        """
        collision_scenes = (
            await self.nova._api_client.store_collision_scenes_api.list_stored_collision_scenes(
                cell=self.nova.cell()._cell_id
            )
        )

        if scene_id not in collision_scenes:
            raise ValueError(f"Collision scene with ID {scene_id} not found")

        log_collision_scenes({scene_id: collision_scenes[scene_id]})
        return {scene_id: collision_scenes[scene_id]}

    def _log_collision_scene(self, collision_scenes: Dict[str, models.CollisionScene]) -> None:
        log_collision_scenes(collision_scenes=collision_scenes)

    async def log_motion(
        self, motion_id: str, timing_mode=TimingMode.CONTINUE, time_offset: float = 0
    ) -> None:
        # Fetch motion details from api
        motion = await self.nova._api_client.motion_api.get_planned_motion(
            self.nova.cell()._cell_id, motion_id
        )
        optimizer_config = (
            await self.nova._api_client.motion_group_infos_api.get_optimizer_configuration(
                self.nova.cell()._cell_id, motion.motion_group
            )
        )
        trajectory = await self.nova._api_client.motion_api.get_motion_trajectory(
            self.nova.cell()._cell_id, motion_id, int(RECORDING_INTERVAL * 1000)
        )

        motion_groups = await self.nova._api_client.motion_group_api.list_motion_groups(
            self.nova.cell()._cell_id
        )
        motion_motion_group = next(
            (mg for mg in motion_groups.instances if mg.motion_group == motion.motion_group), None
        )

        collision_scenes = (
            await self.nova._api_client.store_collision_scenes_api.list_stored_collision_scenes(
                cell=self.nova.cell()._cell_id
            )
        )

        log_motion(
            motion_id=motion_id,
            model_from_controller=motion_motion_group.model_from_controller,
            motion_group=motion.motion_group,
            optimizer_config=optimizer_config,
            trajectory=trajectory.trajectory,
            collision_scenes=collision_scenes,
            time_offset=time_offset,
            timing_mode=timing_mode,
        )

    async def log_trajectory(
        self,
        joint_trajectory: models.JointTrajectory,
        tcp: str,
        motion_group: MotionGroup,
        timing_mode=TimingMode.CONTINUE,
        time_offset: float = 0,
    ) -> None:
        load_plan_response = await motion_group._load_planned_motion(joint_trajectory, tcp)
        await self.log_motion(
            load_plan_response.motion, timing_mode=timing_mode, time_offset=time_offset
        )

    async def log_error_feedback(
        self, error_feedback: PlanTrajectoryFailedResponseErrorFeedback
    ) -> None:
        if isinstance(error_feedback.actual_instance, FeedbackOutOfWorkspace):
            rr.log(
                "motion/errors/FeedbackOutOfWorkspace",
                rr.Points3D(
                    [
                        error_feedback.actual_instance.invalid_tcp_pose.position[0],
                        error_feedback.actual_instance.invalid_tcp_pose.position[1],
                        error_feedback.actual_instance.invalid_tcp_pose.position[2],
                    ],
                    radii=rr.Radius.ui_points([5.0]),
                    colors=[(255, 0, 0, 255)],
                    labels=["Out of Workspace"],
                ),
                timeless=True,
            )

    async def start_streaming(self, motion_group: MotionGroup) -> None:
        """Start streaming real-time robot state to Rerun viewer."""
        if motion_group in self._streaming_tasks:
            return

        task = asyncio.create_task(stream_motion_group(self, motion_group=motion_group))
        self._streaming_tasks[motion_group] = task

    async def stop_streaming(self) -> None:
        """Stop all streaming tasks."""
        for task in self._streaming_tasks.values():
            task.cancel()
        self._streaming_tasks.clear()

    async def log_actions(
        self, actions: list[Action] | Action, show_connection: bool = False
    ) -> None:
        from nova_rerun_bridge import trajectory

        rr.set_time_seconds(TIME_INTERVAL_NAME, trajectory._last_end_time)

        if not isinstance(actions, list):
            actions = [actions]

        if len(actions) == 0:
            raise ValueError("No actions provided")

        poses = CombinedActions(items=tuple(actions)).poses()
        positions = []

        # Collect all positions
        for pose in poses:
            logger.debug(f"Pose: {pose}")
            positions.append([pose.position.x, pose.position.y, pose.position.z])

        # Log all positions at once
        rr.log(
            "motion/actions",
            rr.Points3D(
                positions,
                colors=[colors.colors[0]] * len(positions),  # Green points
                radii=rr.Radius.ui_points([5.0]),
            ),
            timeless=True,
            static=True,
        )

        if show_connection:
            rr.log(
                "motion/actions/connection",
                rr.LineStrips3D([positions], colors=[155, 155, 155, 50]),
            )

    async def __aenter__(self) -> "NovaRerunBridge":
        """Context manager entry point.

        Returns:
            NovaRerunBridge: Self reference for context manager usage.
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Context manager exit point, ensures cleanup."""
        await self.cleanup()

    async def cleanup(self) -> None:
        """Cleanup resources and close Nova API client connection."""
        if hasattr(self.nova, "_api_client"):
            await self.nova._api_client.close()



import asyncio

from nova import Controller, Nova
from nova.actions import jnt, ptp
from numpy import pi

from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode

"""
Example: Show an external axis with a robot moving in a cell.

Prerequisites:
- A cell with a robot with an external axis: setup a yaskawa-ar1440 and import the yaskawa-ar1440-with-external-axis.yaml in the settings app
"""


async def move_robot(controller: Controller, bridge: NovaRerunBridge):
    async with controller[0] as motion_group:
        home_joints = await motion_group.joints()
        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]

        current_pose = await motion_group.tcp_pose(tcp)
        target_pose = current_pose @ (100, 0, 0, 0, 0, 0)
        actions = [jnt(home_joints), ptp(target_pose), jnt(home_joints)]

        trajectory = await motion_group.plan(actions, tcp)
        await bridge.log_trajectory(trajectory, tcp, motion_group, timing_mode=TimingMode.SYNC)


async def move_axis(controller: Controller, bridge: NovaRerunBridge):
    async with controller[16] as motion_group:
        actions = [jnt([0, 0]), jnt([pi / 4, pi]), jnt([0, 0])]

        trajectory = await motion_group.plan(actions, "")
        await bridge.log_trajectory(trajectory, "", motion_group, timing_mode=TimingMode.SYNC)


async def main():
    nova = Nova()
    bridge = NovaRerunBridge(nova)
    await bridge.setup_blueprint()
    cell = nova.cell()
    yaskawa = await cell.controller("yaskawa")
    await asyncio.gather(move_robot(yaskawa, bridge=bridge), move_axis(yaskawa, bridge=bridge))
    await nova.close()
    await bridge.cleanup()


if __name__ == "__main__":
    asyncio.run(main())



from typing import List

import rerun as rr
import rerun.blueprint as rrb

from nova_rerun_bridge import colors
from nova_rerun_bridge.consts import TIME_INTERVAL_NAME


def configure_joint_line_colors(motion_group: str):
    """
    Log the visualization lines for joint limit boundaries.
    """

    for i in range(1, 7):
        prefix = f"motion/{motion_group}/joint"
        color = colors.colors[i - 1]

        rr.log(
            f"{prefix}_velocity_lower_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_velocity_lower_limit_{i}", width=4),
            timeless=True,
        )
        rr.log(
            f"{prefix}_velocity_upper_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_velocity_upper_limit_{i}", width=4),
            timeless=True,
        )

        rr.log(
            f"{prefix}_acceleration_lower_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_acceleration_lower_limit_{i}", width=4),
            timeless=True,
        )
        rr.log(
            f"{prefix}_acceleration_upper_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_acceleration_upper_limit_{i}", width=4),
            timeless=True,
        )

        rr.log(
            f"{prefix}_position_lower_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_position_lower_limit_{i}", width=4),
            timeless=True,
        )
        rr.log(
            f"{prefix}_position_upper_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_position_upper_limit_{i}", width=4),
            timeless=True,
        )

        rr.log(
            f"{prefix}_torque_limit_{i}",
            rr.SeriesLine(color=[176, 49, 40], name=f"joint_torques_lower_limit_{i}", width=4),
            timeless=True,
        )

    for i in range(1, 7):
        prefix = f"motion/{motion_group}/joint"
        color = colors.colors[i - 1]

        rr.log(
            f"{prefix}_velocity_{i}",
            rr.SeriesLine(color=color, name=f"joint_velocity_{i}", width=2),
            timeless=True,
        )
        rr.log(
            f"{prefix}_velocity_{i}",
            rr.SeriesLine(color=color, name=f"joint_velocity_{i}", width=2),
            timeless=True,
        )

        rr.log(
            f"{prefix}_acceleration_{i}",
            rr.SeriesLine(color=color, name=f"joint_acceleration_{i}", width=2),
            timeless=True,
        )
        rr.log(
            f"{prefix}_acceleration_{i}",
            rr.SeriesLine(color=color, name=f"joint_acceleration_{i}", width=2),
            timeless=True,
        )

        rr.log(
            f"{prefix}_position_{i}",
            rr.SeriesLine(color=color, name=f"joint_position_{i}", width=2),
            timeless=True,
        )
        rr.log(
            f"{prefix}_position_{i}",
            rr.SeriesLine(color=color, name=f"joint_position_{i}", width=2),
            timeless=True,
        )

        rr.log(
            f"{prefix}_torque_{i}",
            rr.SeriesLine(color=color, name=f"joint_torques_{i}", width=2),
            timeless=True,
        )


def configure_tcp_line_colors(motion_group: str):
    """
    Configure time series lines for motion data.
    """
    series_specs = [
        ("tcp_velocity", [136, 58, 255], 2),
        ("tcp_acceleration", [136, 58, 255], 2),
        ("tcp_orientation_velocity", [136, 58, 255], 2),
        ("tcp_orientation_acceleration", [136, 58, 255], 2),
        ("time", [136, 58, 255], 2),
        ("location_on_trajectory", [136, 58, 255], 2),
        ("tcp_acceleration_lower_limit", [176, 49, 40], 4),
        ("tcp_acceleration_upper_limit", [176, 49, 40], 4),
        ("tcp_orientation_acceleration_lower_limit", [176, 49, 40], 4),
        ("tcp_orientation_acceleration_upper_limit", [176, 49, 40], 4),
        ("tcp_velocity_limit", [176, 49, 40], 4),
        ("tcp_orientation_velocity_limit", [176, 49, 40], 4),
    ]
    for name, color, width in series_specs:
        rr.log(
            f"motion/{motion_group}/{name}",
            rr.SeriesLine(color=color, name=name, width=width),
            timeless=True,
        )


def joint_content_lists(motion_group: str):
    """
    Generate content lists for joint-related time series.
    """
    velocity_contents = [f"motion/{motion_group}/joint_velocity_{i}" for i in range(1, 7)]
    velocity_limits = [
        f"motion/{motion_group}/joint_velocity_lower_limit_{i}" for i in range(1, 7)
    ] + [f"motion/{motion_group}/joint_velocity_upper_limit_{i}" for i in range(1, 7)]

    accel_contents = [f"motion/{motion_group}/joint_acceleration_{i}" for i in range(1, 7)]
    accel_limits = [
        f"motion/{motion_group}/joint_acceleration_lower_limit_{i}" for i in range(1, 7)
    ] + [f"motion/{motion_group}/joint_acceleration_upper_limit_{i}" for i in range(1, 7)]

    pos_contents = [f"motion/{motion_group}/joint_position_{i}" for i in range(1, 7)]
    pos_limits = [f"motion/{motion_group}/joint_position_lower_limit_{i}" for i in range(1, 7)] + [
        f"motion/{motion_group}/joint_position_upper_limit_{i}" for i in range(1, 7)
    ]

    torque_contents = [f"motion/{motion_group}/joint_torque_{i}" for i in range(1, 7)]
    torque_limits = [f"motion/{motion_group}/joint_torque_limit_{i}" for i in range(1, 7)]

    return (
        velocity_contents,
        velocity_limits,
        accel_contents,
        accel_limits,
        pos_contents,
        pos_limits,
        torque_contents,
        torque_limits,
    )


def create_tcp_tabs(
    motion_group: str, time_ranges: rrb.VisibleTimeRange, plot_legend: rrb.PlotLegend
) -> rrb.Vertical:
    """Create TCP-related time series views."""
    return rrb.Vertical(
        rrb.TimeSeriesView(
            contents=[
                f"motion/{motion_group}/tcp_velocity/**",
                f"motion/{motion_group}/tcp_velocity_limit/**",
            ],
            name="TCP velocity",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        rrb.TimeSeriesView(
            contents=[
                f"motion/{motion_group}/tcp_acceleration/**",
                f"motion/{motion_group}/tcp_acceleration_lower_limit/**",
                f"motion/{motion_group}/tcp_acceleration_upper_limit/**",
            ],
            name="TCP acceleration",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        rrb.TimeSeriesView(
            contents=[
                f"motion/{motion_group}/tcp_orientation_velocity/**",
                f"motion/{motion_group}/tcp_orientation_velocity_limit/**",
            ],
            name="TCP orientation velocity",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        rrb.TimeSeriesView(
            contents=[
                f"motion/{motion_group}/tcp_orientation_acceleration/**",
                f"motion/{motion_group}/tcp_orientation_acceleration_lower_limit/**",
                f"motion/{motion_group}/tcp_orientation_acceleration_upper_limit/**",
            ],
            name="TCP orientation acceleration",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        name="TCP",
    )


def create_joint_tabs(
    motion_group: str, time_ranges: rrb.VisibleTimeRange, plot_legend: rrb.PlotLegend
) -> rrb.Vertical:
    """Create joint-related time series views."""
    (
        velocity_contents,
        velocity_limits,
        accel_contents,
        accel_limits,
        pos_contents,
        pos_limits,
        torque_contents,
        torque_limits,
    ) = joint_content_lists(motion_group)

    return rrb.Vertical(
        rrb.TimeSeriesView(
            contents=velocity_contents + velocity_limits,
            name="Joint velocity",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        rrb.TimeSeriesView(
            contents=accel_contents + accel_limits,
            name="Joint acceleration",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        rrb.TimeSeriesView(
            contents=pos_contents + pos_limits,
            name="Joint position",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        rrb.TimeSeriesView(
            contents=torque_contents + torque_limits,
            name="Joint torque",
            time_ranges=time_ranges,
            plot_legend=plot_legend,
        ),
        name="Joints",
    )


def create_motion_group_tabs(
    motion_group: str, time_ranges: rrb.VisibleTimeRange, plot_legend: rrb.PlotLegend
) -> rrb.Vertical:
    """Create nested tab structure for a motion group."""
    return rrb.Vertical(
        rrb.Tabs(
            create_tcp_tabs(motion_group, time_ranges, plot_legend),
            create_joint_tabs(motion_group, time_ranges, plot_legend),
        ),
        name=f"Motion Group: {motion_group}",
    )


def send_blueprint(motion_group_list: List[str]) -> None:
    """Send blueprint with nested tab structure."""
    for motion_group in motion_group_list:
        configure_tcp_line_colors(motion_group)
        configure_joint_line_colors(motion_group)

    contents = ["motion/**", "collision_scenes/**"] + [f"{group}/**" for group in motion_group_list]

    time_ranges = rrb.VisibleTimeRange(
        TIME_INTERVAL_NAME,
        start=rrb.TimeRangeBoundary.cursor_relative(seconds=-2),
        end=rrb.TimeRangeBoundary.cursor_relative(seconds=2),
    )
    plot_legend = rrb.PlotLegend(visible=False)

    motion_group_tabs = [
        create_motion_group_tabs(group, time_ranges, plot_legend) for group in motion_group_list
    ]

    rr.send_blueprint(
        rrb.Blueprint(
            rrb.Horizontal(
                rrb.Spatial3DView(contents=contents, name="3D Nova", background=[20, 22, 35]),
                rrb.Tabs(
                    *motion_group_tabs,
                    rrb.TextLogView(origin="/logs/motion", name="Motions"),
                    rrb.TextLogView(origin="/logs", name="API Call Logs"),
                ),
                column_shares=[1, 0.3],
            ),
            collapse_panels=True,
        )
    )



from asyncio import run

import numpy as np
import wandelbots_api_client as nova
from decouple import config
from loguru import logger

NOVA_API_HOST = config("NOVA_API", default=None)
NOVA_ACCESS_TOKEN = config("NOVA_ACCESS_TOKEN", default=None)

# get via /api/v1/cells/{cell}
CELL_CONFIG = {
    "controllers": [
        {
            "configuration": {
                "kind": "VirtualController",
                "manufacturer": "universalrobots",
                "position": "[1.17,-1.658,1.405,-1.571,-1.571,1.17,0]",
                "type": "universalrobots-ur5e",
            },
            "name": "ur",
        }
    ],
    "name": "cell",
}


def get_api_client(nova_host: str) -> nova.ApiClient:
    api_base_path = "/api/v1"
    base_url = f"{nova_host}{api_base_path}"
    client_config = nova.Configuration(host=base_url, access_token=NOVA_ACCESS_TOKEN)
    client_config.verify_ssl = False
    return nova.ApiClient(client_config)


async def get_motion_group_infos(
    client: nova.ApiClient, cell_name: str, tcp_id: str
) -> tuple[str, nova.models.OptimizerSetup]:
    controller_api = nova.ControllerApi(client)
    motion_group_api = nova.MotionGroupApi(client)
    motion_group_info_api = nova.MotionGroupInfosApi(client)

    controllers = (await controller_api.list_controllers(cell=cell_name)).instances
    motion_group_instances = await motion_group_api.activate_all_motion_groups(
        cell=cell_name, controller=controllers[0].controller
    )
    motion_group_id = motion_group_instances.instances[0].motion_group
    robot_setup = await motion_group_info_api.get_optimizer_configuration(
        cell=cell_name, motion_group=motion_group_id, tcp=tcp_id
    )

    return motion_group_id, robot_setup


async def build_collision_world(
    client: nova.ApiClient, cell_name: str, robot_setup: nova.models.OptimizerSetup
) -> str:
    collision_api = nova.StoreCollisionComponentsApi(client)
    scene_api = nova.StoreCollisionScenesApi(client)

    # define static colliders, e.g. workpiece
    random_vertices = [1000, 1000, 0] + 1000 * np.random.random((1000, 3))
    collider = nova.models.Collider(
        shape=nova.models.ColliderShape(
            nova.models.ConvexHull2(vertices=random_vertices.tolist(), shape_type="convex_hull")
        )
    )
    await collision_api.store_collider(cell=cell_name, collider="workpiece", collider2=collider)

    # define TCP collider geometry
    tool_collider = nova.models.Collider(
        shape=nova.models.ColliderShape(
            nova.models.Box2(size_x=100, size_y=100, size_z=100, shape_type="box", box_type="FULL")
        )
    )
    await collision_api.store_collision_tool(
        cell=cell_name, tool="tool_box", request_body={"tool_collider": tool_collider}
    )

    # define robot link geometries
    robot_link_colliders = await collision_api.get_default_link_chain(
        cell=cell_name, motion_group_model=robot_setup.motion_group_type
    )
    await collision_api.store_collision_link_chain(
        cell=cell_name, link_chain="robot_links", collider=robot_link_colliders
    )

    # assemble scene
    scene = nova.models.CollisionScene(
        colliders={"workpiece": collider},
        motion_groups={
            "motion_group": nova.models.CollisionMotionGroup(
                tool={"tool_geometry": tool_collider}, link_chain=robot_link_colliders
            )
        },
    )
    scene_id = "collision_scene"
    await scene_api.store_collision_scene(
        cell_name, scene_id, nova.models.CollisionSceneAssembly(scene=scene)
    )
    return scene_id


async def plan_path(
    client: nova.ApiClient, cell_name: str, scene_id: str, robot_setup: nova.models.OptimizerSetup
) -> nova.models.PlanTrajectoryResponse:
    motion_api = nova.MotionApi(client)
    scene_api = nova.StoreCollisionScenesApi(client)
    start_joints = [1.17, -1.658, 1.405, -1.571, -1.571, 1.17]
    goal_pose = [-58.4, -479.4, 500, 2.1397, 2.1397, -0.3571]
    scene = await scene_api.get_stored_collision_scene(cell_name, scene_id)

    plan_request = nova.models.PlanTrajectoryRequest(
        robot_setup=robot_setup,
        start_joint_position=start_joints,
        motion_commands=[
            nova.models.MotionCommand(
                path=nova.models.MotionCommandPath(
                    nova.models.path_line.PathLine(
                        target_pose=nova.models.Pose2(
                            position=goal_pose[:3], orientation=goal_pose[3:]
                        ),
                        path_definition_name="PathLine",
                    )
                ),
                blending=nova.models.MotionCommandBlending(
                    nova.models.motion_command_blending.BlendingPosition(
                        position_zone_radius=20, blending_name="BlendingPosition"
                    )
                ),
            )
        ],
        static_colliders=scene.colliders,
        collision_motion_group=scene.motion_groups.get("motion_group"),
    )

    plan_result = await motion_api.plan_trajectory(
        cell=cell_name, plan_trajectory_request=plan_request
    )
    if isinstance(plan_result.response.actual_instance, nova.models.JointTrajectory):
        logger.info("Plan request successful.")
    else:
        error = plan_result.response.actual_instance.error_feedback.actual_instance
        logger.error(error.error_feedback_name)
        logger.error(error)

    return plan_result


async def upload_planned_motion(
    client: nova.ApiClient,
    plan_response: nova.models.PlanTrajectoryResponse,
    cell_name: str,
    motion_group_id: str,
    tcp_id: str,
) -> str:
    motion_api = nova.MotionApi(api_client=client)
    trajectory = plan_response.response.actual_instance
    planned_motion = nova.models.PlannedMotion(
        motion_group=motion_group_id,
        tcp=tcp_id,
        times=trajectory.times,
        joint_positions=[nova.models.Joints(joints=j.joints) for j in trajectory.joint_positions],
    )
    plan_response = await motion_api.load_planned_motion(
        cell=cell_name, planned_motion=planned_motion
    )

    if plan_response.plan_successful_response is not None:
        motion_id = plan_response.plan_successful_response.motion
        logger.info("Planning was successful. Motion can be executed.")
    else:
        motion_id = plan_response.plan_failed_on_trajectory_response.motion
        logger.warning("Planning failed. Motion can partly be executed.")
        logger.warning(plan_response.plan_failed_on_trajectory_response.description)

    return motion_id


async def execute_motion(client: nova.ApiClient, cell_name: str, motion_id: str):
    motion_api = nova.MotionApi(api_client=client)

    # Move to first point on trajectory
    movement_stream = motion_api.stream_move_to_trajectory_via_joint_ptp(
        cell=cell_name, motion=motion_id, location_on_trajectory=0
    )
    async for m in movement_stream:
        if m.move_response:
            continue
        logger.info(f"{m.stop_response.location_on_trajectory=}")

    # Move along trajectory
    movement_stream = motion_api.stream_move_forward(
        cell=cell_name, motion=motion_id, playback_speed_in_percent=100
    )
    async for m in movement_stream:
        if m.move_response:
            continue
        logger.info(f"{m.stop_response.location_on_trajectory=}")


async def main():
    tcp_id = "Flange"
    cell_name = CELL_CONFIG["name"]

    client = get_api_client(NOVA_API_HOST)
    motion_group_id, robot_setup = await get_motion_group_infos(client, cell_name, tcp_id)
    scene_id = await build_collision_world(client, cell_name, robot_setup)
    plan_result = await plan_path(client, cell_name, scene_id, robot_setup)
    motion_id = await upload_planned_motion(client, plan_result, cell_name, motion_group_id, tcp_id)
    await execute_motion(client, cell_name, motion_id)

    await client.close()


if __name__ == "__main__":
    run(main())



from typing import Any, List, Tuple

import numpy as np
import trimesh
from scipy.spatial import ConvexHull


class HullVisualizer:
    @staticmethod
    def compute_hull_mesh(
        polygons: List[np.ndarray],
    ) -> Tuple[List[List[float]], List[List[int]], List[List[float]]]:
        """Convert polygons to mesh with optimized hull generation."""
        vertices = np.vstack(polygons)

        # Custom qhull options for better quality
        qhull_opts = trimesh.convex.QhullOptions(
            Qt=True,  # Triangulated output
            QJ=True,  # Joggled input for precision
            Qc=True,  # Keep coplanar points
            Qx=True,  # Exact pre-merges
            QbB=True,  # Scale to unit cube
            Pp=True,  # Remove precision warnings
        )

        mesh = trimesh.convex.convex_hull(vertices, qhull_options=qhull_opts, repair=True)

        return (mesh.vertices.tolist(), mesh.faces.tolist(), mesh.vertex_normals.tolist())

    @staticmethod
    def plane_from_triangle(p0, p1, p2, normal_epsilon=1e-6):
        # Compute normal
        v1 = p1 - p0
        v2 = p2 - p0
        n = np.cross(v1, v2)
        norm = np.linalg.norm(n)
        if norm < normal_epsilon:
            return None, None
        n = n / norm
        # Plane: n·x = d
        d = np.dot(n, p0)
        return n, d

    @staticmethod
    def group_coplanar_triangles(points, hull, angle_epsilon=1e-6, dist_epsilon=1e-6):
        # Group triangles by their plane (normal and distance)
        plane_map = {}
        for simplex in hull.simplices:
            p0, p1, p2 = points[simplex]
            n, d = HullVisualizer.plane_from_triangle(p0, p1, p2)
            if n is None:
                continue

            # Ensure a canonical representation of the plane normal
            for i_comp in range(3):
                if abs(n[i_comp]) > angle_epsilon:
                    if n[i_comp] < 0:
                        n = -n
                        d = -d
                    break

            # Round normal and distance for stable hashing
            n_rounded = tuple(np.round(n, 6))
            d_rounded = round(d, 3)

            key = (n_rounded, d_rounded)
            if key not in plane_map:
                plane_map[key] = []
            plane_map[key].append(simplex)

        return plane_map

    @staticmethod
    def merge_coplanar_triangles_to_polygon(points, simplices):
        # Extract polygon outline from coplanar triangles
        edges = {}
        for tri in simplices:
            for i in range(3):
                a = tri[i]
                b = tri[(i + 1) % 3]
                e = (min(a, b), max(a, b))
                edges[e] = edges.get(e, 0) + 1

        # Keep only outer edges (appear once)
        boundary_edges = [e for e, count in edges.items() if count == 1]
        if not boundary_edges:
            return []

        # Build adjacency
        adj = {}
        for a, b in boundary_edges:
            adj.setdefault(a, []).append(b)
            adj.setdefault(b, []).append(a)

        # Walk along the boundary edges to form a closed loop
        start = boundary_edges[0][0]
        loop = [start]
        current = start
        prev = None
        while True:
            neighbors = adj[current]
            next_vertex = None
            for n in neighbors:
                if n != prev:
                    next_vertex = n
                    break
            if next_vertex is None:
                break
            loop.append(next_vertex)
            prev, current = current, next_vertex
            if next_vertex == start:
                break

        polygon_points = points[loop]
        return polygon_points

    @staticmethod
    def compute_hull_outlines_from_geometries(child_geometries: List[Any]) -> List[np.ndarray]:
        """Compute polygon outlines from geometry child objects.

        Args:
            child_geometries: List of geometry objects containing convex hulls

        Returns:
            List of closed polygons as Nx3 numpy arrays
        """
        all_points = []
        for child in child_geometries:
            if child.convex_hull is not None:
                for v in child.convex_hull.vertices:
                    all_points.append([v.x, v.y, v.z])

        if len(all_points) < 4:
            return []

        return HullVisualizer._compute_hull_from_points(np.array(all_points))

    @staticmethod
    def compute_hull_outlines_from_points(points: List[List[float]]) -> List[np.ndarray]:
        """Compute polygon outlines directly from point coordinates.

        Args:
            points: List of [x,y,z] coordinates

        Returns:
            List of closed polygons as Nx3 numpy arrays
        """
        if len(points) < 4:
            return []

        return HullVisualizer._compute_hull_from_points(np.array(points))

    @staticmethod
    def _compute_hull_from_points(points: np.ndarray) -> List[np.ndarray]:
        """Internal helper to compute hull from numpy points array."""
        try:
            hull = ConvexHull(points)
            plane_map = HullVisualizer.group_coplanar_triangles(points, hull)

            polygons = []
            for simplices in plane_map.values():
                polygon_points = HullVisualizer.merge_coplanar_triangles_to_polygon(
                    points, simplices
                )
                if len(polygon_points) > 2:
                    closed_loop = np.vstack([polygon_points, polygon_points[0]])
                    polygons.append(closed_loop)
            return polygons

        except Exception:
            return []



import json
import os

PROCESSED_MOTIONS_FILE = "processed_motions.json"


def load_processed_motions():
    if os.path.exists(PROCESSED_MOTIONS_FILE):
        with open(PROCESSED_MOTIONS_FILE, "r") as file:
            return set(tuple(item) for item in json.load(file))
    return set()


def save_processed_motion(motion_id, trajectory_time):
    processed_motions = load_processed_motions()
    processed_motions.add((motion_id, trajectory_time))
    with open(PROCESSED_MOTIONS_FILE, "w") as file:
        json.dump(list(processed_motions), file)



from typing import List

import numpy as np
from nova.api import models


class DHRobot:
    """A class for handling DH parameters and computing joint positions."""

    def __init__(self, dh_parameters: List[models.DHParameter], mounting: models.PlannerPose):
        """
        Initialize the DHRobot with DH parameters and a mounting pose.
        :param dh_parameters: List of DHParameter objects containing all joint configurations.
        :param mounting: PlannerPose object representing the mounting orientation and position.
        """
        self.dh_parameters = dh_parameters
        self.mounting = mounting

    def pose_to_matrix(self, pose: models.PlannerPose):
        """
        Convert a PlannerPose (with quaternion orientation) into a 4x4 homogeneous transformation matrix.
        :param pose: A PlannerPose object with position: Vector3d and orientation: Quaternion.
        :return: A 4x4 numpy array representing the transformation.
        """
        # Extract translation
        if pose.position is None:
            x, y, z = 0.0, 0.0, 0.0
        else:
            x, y, z = pose.position.x, pose.position.y, pose.position.z

        # Extract quaternion
        if pose.orientation is None:
            # If no orientation is provided, assume identity orientation
            w, qx, qy, qz = 1.0, 0.0, 0.0, 0.0
        else:
            w = pose.orientation.w
            qx = pose.orientation.x
            qy = pose.orientation.y
            qz = pose.orientation.z

        # Compute rotation matrix from quaternion
        # R = [[1 - 2(y²+z²), 2(xy - zw),     2(xz + yw)    ],
        #      [2(xy + zw),     1 - 2(x²+z²), 2(yz - xw)    ],
        #      [2(xz - yw),     2(yz + xw),   1 - 2(x²+y²)]]

        xx = qx * qx
        yy = qy * qy
        zz = qz * qz
        xy = qx * qy
        xz = qx * qz
        yz = qy * qz
        wx = w * qx
        wy = w * qy
        wz = w * qz

        R = np.array(
            [
                [1.0 - 2.0 * (yy + zz), 2.0 * (xy - wz), 2.0 * (xz + wy)],
                [2.0 * (xy + wz), 1.0 - 2.0 * (xx + zz), 2.0 * (yz - wx)],
                [2.0 * (xz - wy), 2.0 * (yz + wx), 1.0 - 2.0 * (xx + yy)],
            ]
        )

        # Construct the full homogeneous transformation matrix
        T = np.eye(4)
        T[0:3, 0:3] = R
        T[0:3, 3] = [x, y, z]

        return T

    def dh_transform(self, dh_param: models.DHParameter, joint_rotation):
        """
        Compute the homogeneous transformation matrix for a given DH parameter and joint rotation.
        :param dh_param: A single DH parameter.
        :param joint_rotation: The joint rotation value (in radians).
        :return: A 4x4 homogeneous transformation matrix.
        """
        # Adjust the angle based on rotation direction
        theta = dh_param.theta + joint_rotation * (-1 if dh_param.reverse_rotation_direction else 1)
        d = dh_param.d
        a = dh_param.a
        alpha = dh_param.alpha

        # Create the homogeneous transformation matrix for this DH parameter
        transformation = np.array(
            [
                [
                    np.cos(theta),
                    -np.sin(theta) * np.cos(alpha),
                    np.sin(theta) * np.sin(alpha),
                    a * np.cos(theta),
                ],
                [
                    np.sin(theta),
                    np.cos(theta) * np.cos(alpha),
                    -np.cos(theta) * np.sin(alpha),
                    a * np.sin(theta),
                ],
                [0, np.sin(alpha), np.cos(alpha), d],
                [0, 0, 0, 1],
            ]
        )
        return transformation

    def calculate_joint_positions(self, joint_values):
        """
        Compute joint positions based on joint values.
        :param joint_values: Object containing joint rotation values as a list in joint_values.joints.
        :return: A list of joint positions as [x, y, z].
        """
        # Incorporate the mounting pose at the start
        accumulated_matrix = self.pose_to_matrix(self.mounting)

        joint_positions = [
            accumulated_matrix[:3, 3].tolist()
        ]  # Base position after mounting is applied

        for dh_param, joint_rotation in zip(self.dh_parameters, joint_values.joints):
            transform = self.dh_transform(dh_param, joint_rotation)
            accumulated_matrix = accumulated_matrix @ transform
            position = accumulated_matrix[:3, 3]  # Extract translation (x, y, z)
            joint_positions.append(position.tolist())

        return joint_positions
